@@ -12,6 +12,18 @@ import { CursorIDEBrowserLogger } from './logger.js';
 const SNAPSHOT_SIZE_THRESHOLD = 25 * 1024; // 25KB threshold for snapshots
 const SNAPSHOT_PREVIEW_LINES = 50; // Number of lines to preview
 const TEMP_LOG_DIR = path.join(os.homedir(), '.cursor', 'browser-logs');
+const ORIGIN_NAVIGATION_COMMAND = 'cursor.browserOriginAllowlist.ensureNavigationAllowed';
+const ORIGIN_PAGE_COMMAND = 'cursor.browserOriginAllowlist.ensurePageOriginAllowed';
+const BROWSER_GET_URL_COMMAND = 'cursor.browserView.getURL';
+
+async function ensureBrowserNavigationAllowed(url: string): Promise<void> {
+	await vscode.commands.executeCommand(ORIGIN_NAVIGATION_COMMAND, { url });
+}
+
+async function ensureCurrentPageAllowed(toolName: string): Promise<void> {
+	const currentUrl = await vscode.commands.executeCommand<string | undefined>(BROWSER_GET_URL_COMMAND);
+	await vscode.commands.executeCommand(ORIGIN_PAGE_COMMAND, { toolName, url: currentUrl });
+}
 
 /**
  * Convert accessibility tree object to YAML-like format string
@@ -497,26 +509,131 @@ async function executeInBrowser(code: string): Promise<unknown> {
 	}
 }
 
+interface PageState {
+	url: string;
+	title: string;
+	snapshot: unknown;
+}
+
+interface FormatPageStateOptions {
+	/** The action name (e.g., 'click', 'type', 'hover') */
+	action: string;
+	/** Human-readable detail lines to include in the output */
+	detailLines?: string[];
+	/** The page state containing url, title, and snapshot */
+	pageState: PageState;
+}
+
+/**
+ * Format a browser action result with properly formatted YAML snapshot
+ */
+async function formatPageStateResult(options: FormatPageStateOptions): Promise<McpToolResult> {
+	const { action, detailLines, pageState } = options;
+
+	if (pageState?.snapshot) {
+		// Convert the accessibility tree object to YAML format string
+		const ariaSnapshot = accessibilityTreeToYaml(pageState.snapshot);
+		const url = pageState.url;
+		const title = pageState.title;
+
+		// Check if snapshot is too large and should be written to file
+		const snapshotSize = Buffer.byteLength(ariaSnapshot, 'utf8');
+		if (snapshotSize > SNAPSHOT_SIZE_THRESHOLD) {
+			try {
+				const { filePath, previewLines, totalLines } = await writeSnapshotToFile(ariaSnapshot);
+
+				// Return text-based result with file reference instead of full snapshot
+				const textParts: string[] = [];
+				textParts.push(`### Action: ${action}`);
+				if (detailLines && detailLines.length > 0) {
+					textParts.push(...detailLines);
+				}
+				textParts.push(
+					`\n### Page state`,
+					`- Page URL: ${url}`,
+					`- Page Title: ${title}`,
+					`- Page Snapshot: Large snapshot (${snapshotSize} bytes, ${totalLines} lines) written to file`,
+					`- Snapshot File: [${filePath}](file://${filePath})`,
+					`- Preview (first ${previewLines.length} lines):`,
+					'```yaml',
+					previewLines.join('\n'),
+					'```',
+					`\n... (${totalLines - previewLines.length} more lines in file)`
+				);
+
+				return {
+					content: [{ type: 'text', text: textParts.join('\n') }]
+				};
+			} catch (error) {
+				CursorIDEBrowserLogger.error('Failed to write snapshot to file, returning inline:', error);
+				// Fall back to returning the full snapshot inline
+			}
+		}
+
+		// Return inline snapshot for small snapshots or fallback
+		const textParts: string[] = [];
+		textParts.push(`### Action: ${action}`);
+		if (detailLines && detailLines.length > 0) {
+			textParts.push(...detailLines);
+		}
+		textParts.push(
+			`\n### Page state`,
+			`- Page URL: ${url}`,
+			`- Page Title: ${title}`,
+			`- Page Snapshot:`,
+			'```yaml',
+			ariaSnapshot,
+			'```'
+		);
+
+		return {
+			content: [{ type: 'text', text: textParts.join('\n') }]
+		};
+	}
+
+	// If we reach this point, for some reason there's no snapshot in the pageState
+	// so just return whatever is in the result
+	return {
+		content: [{ type: 'text', text: JSON.stringify(options) }]
+	};
+}
+
+export interface McpToolResultContentItem {
+	type: 'text' | 'image';
+	text?: string;
+	data?: string;
+	mimeType?: string;
+}
+
+// TODO: Replace with the proto
+export interface McpToolResult {
+	content: McpToolResultContentItem[];
+}
+
 /**
  * Browser tool implementations using direct JavaScript execution
  */
 export const BrowserTools = {
-	async navigate(params: { url: string }): Promise<unknown> {
+	async navigate(params: { url: string }): Promise<McpToolResult> {
+		await ensureBrowserNavigationAllowed(params.url);
 		CursorIDEBrowserLogger.info(`Navigating to ${params.url}`);
-		await vscode.commands.executeCommand('cursor.browserView.navigate', params.url);
-
+		await vscode.commands.executeCommand('workbench.action.openBrowserEditor', {
+			preserveFocus: true,
+			url: params.url,
+			trackingId: 'mcp_tool',
+		});
 		// Wait for page to load and return snapshot
 		await new Promise(resolve => setTimeout(resolve, 1000));
 		return await BrowserTools.snapshot({});
 	},
 
-	async snapshot(_params: Record<string, unknown>): Promise<unknown> {
+	async snapshot(_params: Record<string, unknown>): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_snapshot');
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
 				const tree = buildPageSnapshot();
 				return {
-					action: 'snapshot',
 					success: true,
 					pageState: {
 						url: window.location.href,
@@ -526,70 +643,19 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as any;
+		const result = await executeInBrowser(code) as {
+			success: boolean;
+			pageState: PageState;
+		};
 
-		// Check if the snapshot is large and should be written to a file
-		if (result?.pageState?.snapshot) {
-			// Convert the accessibility tree object to YAML format string
-			const ariaSnapshot = accessibilityTreeToYaml(result.pageState.snapshot);
-			const url = result.pageState.url;
-			const title = result.pageState.title;
-
-			// Check if snapshot is too large and should be written to file
-			const snapshotSize = Buffer.byteLength(ariaSnapshot, 'utf8');
-			if (snapshotSize > SNAPSHOT_SIZE_THRESHOLD) {
-				try {
-					const { filePath, previewLines, totalLines } = await writeSnapshotToFile(ariaSnapshot);
-
-					// Return text-based result with file reference instead of full snapshot
-					const textParts: string[] = [];
-					textParts.push(
-						`\n\n### Page state`,
-						`- Page URL: ${url}`,
-						`- Page Title: ${title}`,
-						`- Page Snapshot: Large snapshot (${snapshotSize} bytes, ${totalLines} lines) written to file`,
-						`- Snapshot File: [${filePath}](file://${filePath})`,
-						`- Preview (first ${previewLines.length} lines):`,
-						'```yaml',
-						previewLines.join('\n'),
-						'```',
-						`\n... (${totalLines - previewLines.length} more lines in file)`
-					);
-
-					return {
-						action: 'snapshot',
-						success: true,
-						content: [{ type: 'text', text: textParts.join('\n') }]
-					};
-				} catch (error) {
-					CursorIDEBrowserLogger.error('Failed to write snapshot to file, returning inline:', error);
-					// Fall back to returning the full snapshot inline
-				}
-			}
-
-			// Return inline snapshot for small snapshots or fallback
-			const textParts: string[] = [];
-			textParts.push(
-				`\n\n### Page state`,
-				`- Page URL: ${url}`,
-				`- Page Title: ${title}`,
-				`- Page Snapshot:`,
-				'```yaml',
-				ariaSnapshot,
-				'```'
-			);
-
-			return {
-				action: 'snapshot',
-				success: true,
-				content: [{ type: 'text', text: textParts.join('\n') }]
-			};
-		}
-
-		return result;
+		return await formatPageStateResult({
+			action: 'snapshot',
+			pageState: result.pageState,
+		});
 	},
 
-	async click(params: { ref: string; element?: string; doubleClick?: boolean; button?: string; modifiers?: string[] }): Promise<unknown> {
+	async click(params: { ref: string; element?: string; doubleClick?: boolean; button?: string; modifiers?: string[] }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_click');
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -639,12 +705,9 @@ export const BrowserTools = {
 
 				const snapshot = buildPageSnapshot();
 				return {
-					action: 'click',
 					success: true,
-					details: {
-						doubleClick: ${JSON.stringify(params.doubleClick)} || false,
-						button: ${JSON.stringify(params.button)} || 'left'
-					},
+					doubleClick: ${JSON.stringify(params.doubleClick)} || false,
+					button: ${JSON.stringify(params.button)} || 'left',
 					pageState: {
 						url: window.location.href,
 						title: document.title,
@@ -653,10 +716,26 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		return await executeInBrowser(code);
+		const result = await executeInBrowser(code) as {
+			success: boolean;
+			doubleClick: boolean;
+			button: string;
+			pageState: PageState;
+		};
+
+		const detailLines: string[] = [];
+		detailLines.push(`- Click type: ${result.doubleClick ? 'double-click' : 'single-click'}`);
+		detailLines.push(`- Button: ${result.button}`);
+
+		return await formatPageStateResult({
+			action: 'click',
+			detailLines,
+			pageState: result.pageState,
+		});
 	},
 
-	async type(params: { ref: string; text: string; element?: string; submit?: boolean; slowly?: boolean }): Promise<unknown> {
+	async type(params: { ref: string; text: string; element?: string; submit?: boolean; slowly?: boolean }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_type');
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -713,9 +792,10 @@ export const BrowserTools = {
 
 				const snapshot = buildPageSnapshot();
 				return {
-					action: 'type',
 					success: true,
-					details: { slowly, submitted: submit },
+					slowly,
+					submitted: submit,
+					textLength: text.length,
 					pageState: {
 						url: window.location.href,
 						title: document.title,
@@ -724,10 +804,30 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		return await executeInBrowser(code);
+		const result = await executeInBrowser(code) as {
+			success: boolean;
+			slowly: boolean;
+			submitted: boolean;
+			textLength: number;
+			pageState: PageState;
+		};
+
+		const detailLines: string[] = [];
+		detailLines.push(`- Characters typed: ${result.textLength}`);
+		detailLines.push(`- Typing mode: ${result.slowly ? 'slow (character-by-character)' : 'fast'}`);
+		if (result.submitted) {
+			detailLines.push(`- Form submitted: yes (Enter key pressed)`);
+		}
+
+		return await formatPageStateResult({
+			action: 'type',
+			detailLines,
+			pageState: result.pageState,
+		});
 	},
 
-	async hover(params: { ref: string; element?: string }): Promise<unknown> {
+	async hover(params: { ref: string; element?: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_hover');
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -751,7 +851,6 @@ export const BrowserTools = {
 
 				const snapshot = buildPageSnapshot();
 				return {
-					action: 'hover',
 					success: true,
 					pageState: {
 						url: window.location.href,
@@ -761,10 +860,19 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		return await executeInBrowser(code);
+		const result = await executeInBrowser(code) as {
+			success: boolean;
+			pageState: PageState;
+		};
+
+		return await formatPageStateResult({
+			action: 'hover',
+			pageState: result.pageState,
+		});
 	},
 
-	async selectOption(params: { ref: string; values: string[]; element?: string }): Promise<unknown> {
+	async selectOption(params: { ref: string; values: string[]; element?: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_select_option');
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -804,9 +912,8 @@ export const BrowserTools = {
 
 				const snapshot = buildPageSnapshot();
 				return {
-					action: 'select_option',
 					success: true,
-					details: { selectedValues },
+					selectedValues,
 					pageState: {
 						url: window.location.href,
 						title: document.title,
@@ -815,10 +922,28 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		return await executeInBrowser(code);
+		const result = await executeInBrowser(code) as {
+			success: boolean;
+			selectedValues: string[];
+			pageState: PageState;
+		};
+
+		const detailLines: string[] = [];
+		if (result.selectedValues.length > 0) {
+			detailLines.push(`- Selected values: ${result.selectedValues.join(', ')}`);
+		} else {
+			detailLines.push(`- Selected values: (none)`);
+		}
+
+		return await formatPageStateResult({
+			action: 'select_option',
+			detailLines,
+			pageState: result.pageState,
+		});
 	},
 
-	async pressKey(params: { key: string }): Promise<unknown> {
+	async pressKey(params: { key: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_press_key');
 		const code = `
 			(function() {
 				const key = ${JSON.stringify(params.key)};
@@ -831,13 +956,24 @@ export const BrowserTools = {
 				return { action: 'press_key', success: true, key };
 			})();
 		`;
-		return await executeInBrowser(code);
+		const result = await executeInBrowser(code) as { action: string; success: boolean; key: string };
+		if (!result.success) {
+			return {
+				content: [{ type: 'text', text: `Failed to press key: ${result.key}` }]
+			};
+		}
+		return {
+			content: [{ type: 'text', text: `Successfully pressed key: ${result.key}` }]
+		};
 	},
 
-	async waitFor(params: { time?: number; text?: string; textGone?: string }): Promise<unknown> {
+	async waitFor(params: { time?: number; text?: string; textGone?: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_wait_for');
 		if (params.time !== undefined) {
 			await new Promise(resolve => setTimeout(resolve, params.time! * 1000));
-			return { action: 'wait_for', success: true, type: 'time', duration: params.time };
+			return {
+				content: [{ type: 'text', text: `Successfully waited for: ${params.time} seconds` }]
+			};
 		}
 
 		const code = `
@@ -876,20 +1012,28 @@ export const BrowserTools = {
 				});
 			})();
 		`;
-		return await executeInBrowser(code);
-	},
-
-	async consoleMessages(_params: Record<string, unknown>): Promise<unknown> {
-		// Console logs are now tracked by the main process
-		const logs = await vscode.commands.executeCommand<Array<{ type: string; message: string; timestamp: number }>>('cursor.browserView.getConsoleLogs');
+		const { success, type, text, error } = await executeInBrowser(code) as { action: string; success: boolean; type: string; text?: string; error?: string };
+		if (!success) {
+			return {
+				content: [{ type: 'text', text: `Failed to wait for element: ${error || 'Unknown error'}` }]
+			};
+		}
 		return {
-			action: 'console_messages',
-			success: true,
-			messages: logs
+			content: [{ type: 'text', text: `Successfully waited for element: ${text || 'Unknown text'}` }]
 		};
 	},
 
-	async networkRequests(_params: Record<string, unknown>): Promise<unknown> {
+	async consoleMessages(_params: Record<string, unknown>): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_console_messages');
+		// Console logs are now tracked by the main process
+		const logs = await vscode.commands.executeCommand<Array<{ type: string; message: string; timestamp: number }>>('cursor.browserView.getConsoleLogs');
+		return {
+			content: [{ type: 'text', text: JSON.stringify(logs, null, 2) }]
+		};
+	},
+
+	async networkRequests(_params: Record<string, unknown>): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_network_requests');
 		// Network requests are now tracked by the main process
 		const requests = await vscode.commands.executeCommand<Array<{
 			url: string;
@@ -899,23 +1043,110 @@ export const BrowserTools = {
 			resourceType?: string;
 		}>>('cursor.browserView.getNetworkRequests');
 		return {
-			action: 'network_requests',
-			success: true,
-			requests
+			content: [{ type: 'text', text: JSON.stringify(requests, null, 2) }]
 		};
 	},
 
-	async takeScreenshot(params: { filename?: string; type?: string; fullPage?: boolean; element?: string; ref?: string }): Promise<unknown> {
+	async takeScreenshot(params: { filename?: string; type?: string; fullPage?: boolean; element?: string; ref?: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_take_screenshot');
+
+		// Ensure filename has the correct extension if provided without one
+		let processedParams = params;
+		if (params.filename) {
+			const hasExtension = /\.[a-zA-Z0-9]+$/.test(params.filename);
+			if (!hasExtension) {
+				const fileType = params.type || 'png';
+				processedParams = { ...params, filename: `${params.filename}.${fileType}` };
+			}
+		}
+
 		// Screenshot will be handled by the main process
-		return await vscode.commands.executeCommand('cursor.browserView.takeScreenshot', params);
+		const screenshotResult = await vscode.commands.executeCommand('cursor.browserView.takeScreenshot', processedParams) as {
+			success: boolean;
+			filename?: string;
+			savedPath?: string;
+			dataUrl?: string;
+			error?: string;
+		};
+
+		// If there's an error, return text content with the error
+		if (!screenshotResult.success || screenshotResult.error) {
+			return {
+				content: [{
+					type: 'text',
+					text: `Screenshot failed: ${screenshotResult.error || 'Unknown error'}`
+				}]
+			};
+		}
+
+		// Extract base64 data from data URL
+		if (screenshotResult.dataUrl) {
+			// dataUrl format: "data:image/png;base64,iVBORw0KG..."
+			const match = screenshotResult.dataUrl.match(/^data:([^;]+);base64,(.+)$/);
+			if (match) {
+				const mimeType = match[1];
+				const base64Data = match[2];
+
+				// Build text description
+				const textParts = ['Screenshot captured successfully'];
+				if (screenshotResult.filename) {
+					textParts.push(`Filename: ${screenshotResult.filename}`);
+				}
+				if (screenshotResult.savedPath) {
+					textParts.push(`Saved to: ${screenshotResult.savedPath}`);
+				}
+
+				return {
+					content: [
+						{
+							type: 'text',
+							text: textParts.join('\n')
+						},
+						{
+							type: 'image',
+							data: base64Data,
+							mimeType: mimeType
+						}
+					]
+				};
+			}
+		}
+
+		// Fallback if dataUrl is missing or malformed
+		return {
+			content: [{
+				type: 'text',
+				text: screenshotResult.savedPath
+					? `Screenshot saved to: ${screenshotResult.savedPath}`
+					: 'Screenshot captured but no image data available'
+			}]
+		};
 	},
 
-	async goBack(_params: Record<string, unknown>): Promise<unknown> {
-		return await vscode.commands.executeCommand('cursor.browserView.goBack');
+	async goBack(_params: Record<string, unknown>): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_navigate_back');
+		const success = await vscode.commands.executeCommand('cursor.browserView.goBack') as boolean;
+		if (!success) {
+			return {
+				content: [{ type: 'text', text: 'Failed to navigate back' }]
+			};
+		}
+		return {
+			content: [{ type: 'text', text: 'Successfully navigated back' }]
+		};
 	},
 
-	async resize(params: { width: number; height: number }): Promise<unknown> {
-		return await vscode.commands.executeCommand('cursor.browserView.resize', params);
+	async resize(params: { width: number; height: number }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_resize');
+		const { success } = await vscode.commands.executeCommand('cursor.browserView.resize', params) as { success: boolean };
+		if (!success) {
+			return {
+				content: [{ type: 'text', text: 'Failed to resize browser' }]
+			};
+		}
+		return {
+			content: [{ type: 'text', text: 'Successfully resized browser' }]
+		};
 	}
 };
 
